from django.contrib import messages
from django.contrib.auth.views import REDIRECT_FIELD_NAME, login
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache

from .forms import LoginForm


class CruditorMixin(object):
    menu_title = 'CRUDitor'
    index_url = '#'
    logout_url = '#'
    change_password_url = None
    menu_template_name = 'cruditor/includes/menu.html'
    extrahead_template_name = 'cruditor/includes/extrahead.html'
    login_template_name = 'cruditor/login.html'
    staff_required = True
    required_permission = None

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_active or (self.staff_required and not request.user.is_staff):
            login_defaults = {
                'template_name': self.login_template_name,
                'authentication_form': LoginForm,
                'extra_context': {
                    'app_path': request.get_full_path(),
                    'next_field': REDIRECT_FIELD_NAME,
                    'next_value': request.get_full_path(),
                    'cruditor': self.get_cruditor_context(alternative_title='Login'),
                },
            }
            return login(request, **login_defaults)

        self.ensure_required_permission()

        return super(CruditorMixin, self).dispatch(request, *args, **kwargs)

    def get_cruditor_context(self, alternative_title=None):
        return {
            'title': alternative_title or self.get_title(),
            'breadcrumb': self.get_breadcrumb() + [
                {'title': alternative_title or self.get_breadcrumb_title(), 'url': None}
            ],
            'constants': {
                'menu_title': self.menu_title,
                'menu_template_name': self.menu_template_name,
                'extrahead_template_name': self.extrahead_template_name,
                'index_url': self.index_url,
                'logout_url': self.logout_url,
                'change_password_url': self.change_password_url,
            }
        }

    def get_title(self):
        return getattr(self, 'title', self.menu_title)

    def get_breadcrumb_title(self):
        return self.get_title()

    def get_breadcrumb(self):
        return []

    def get_required_permission(self):
        return self.required_permission

    def ensure_required_permission(self):
        required_permission = self.get_required_permission()

        if not required_permission:
            return

        if not self.request.user.has_perm(required_permission):
            raise PermissionDenied

    def get_context_data(self, **kwargs):
        context = super(CruditorMixin, self).get_context_data(**kwargs)
        context['cruditor'] = self.get_cruditor_context()
        return context


class FormsetViewMixin(object):
    formset_classes = None

    def get_formset_classes(self):
        return self.formset_classes or {}

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        return self.render_to_response(self.get_context_data(
            form=self.get_form(self.get_form_class()),
            **dict([(
                formset_name,
                formset_class(instance=self.object)
            ) for formset_name, formset_class in self.get_formset_classes().items()])
        ))

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()

        form = self.get_form(self.get_form_class())
        formsets = dict([(
            formset_name,
            formset_class(request.POST, files=request.FILES, instance=self.object)
        ) for formset_name, formset_class in self.get_formset_classes().items()])

        if all([form.is_valid()] + [formset.is_valid() for formset in formsets.values()]):
            return self.form_valid(form, **formsets)
        else:
            return self.form_invalid(form, **formsets)

    def form_valid(self, form, **formsets):
        self.object = form.save()

        self.formset_objects = {}
        for formset_name, formset in formsets.items():
            formset.instance = self.object
            self.formset_objects[formset_name] = formset.save()

        messages.success(self.request, self.success_message.format(
            model=self.model._meta.verbose_name, object=self.object))

        return redirect(self.get_success_url())

    def form_invalid(self, form, **formsets):
        return self.render_to_response(self.get_context_data(
            form=form,
            formset_errors=True,
            **formsets
        ))



import functools
import operator

from django.db.models import Q
from django.utils.translation import ugettext
from django_filters import CharFilter, ChoiceFilter


class AnyChoiceFilter(ChoiceFilter):

    def __init__(self, *args, **kwargs):
        empty_label = kwargs.pop('empty_label', ugettext('Any choice'))
        super(AnyChoiceFilter, self).__init__(*args, **kwargs)
        self.extra['choices'] = [('', empty_label)] + list(self.extra['choices'])


class MultiFieldFilter(CharFilter):
    """
    This filter performs an OR query on the defined fields from a
    single entered value.

    The following will work similar to the default UserAdmin search::

        class UserFilterSet(FilterSet):
            search = MultiFieldFilter([
                'username', 'first_name', 'last_name', 'email'])

            class Meta:
                model = User
                fields = ['search']

    Based on https://gist.github.com/nkryptic/4727865
    """

    def __init__(self, fields, *args, **kwargs):
        super(MultiFieldFilter, self).__init__(*args, **kwargs)
        self.fields = fields
        self.lookup_type = 'icontains'
        self.lookup_types = [
            ('^', 'istartswith'),
            ('=', 'iexact'),
            ('@', 'search'),
        ]

    def filter(self, qs, value):
        if not self.fields or not value:
            return qs

        lookups = [self._get_lookup(str(field)) for field in self.fields]
        queries = [Q(**{lookup: value}) for lookup in lookups]
        qs = qs.filter(functools.reduce(operator.or_, queries))
        return qs

    def _get_lookup(self, field_name):
        for key, lookup_type in self.lookup_types:
            if field_name.startswith(key):
                return '{0}__{1}'.format(field_name[len(key):], lookup_type)
        return '{0}__{1}'.format(field_name, self.lookup_type)



import codecs
import os
import sys

from setuptools import setup, find_packages


VERSION = __import__('cruditor').__version__


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    print('You probably want to also tag the version now:')
    print('  git tag -a %s -m "version %s"' % (VERSION, VERSION))
    print('  git push --tags')
    sys.exit()


def read(*parts):
    filename = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(filename, encoding='utf-8') as fp:
        return fp.read()


# TEMPORARY FIX FOR
# https://bitbucket.org/pypa/setuptools/issues/450/egg_info-command-is-very-slow-if-there-are
TO_OMIT = ['.git', '.tox']
orig_os_walk = os.walk
def patched_os_walk(path, *args, **kwargs):
    for (dirpath, dirnames, filenames) in orig_os_walk(path, *args, **kwargs):
        if '.git' in dirnames:
            # We're probably in our own root directory.
            print("MONKEY PATCH: omitting a few directories like .git and .tox...")
            dirnames[:] = list(set(dirnames) - set(TO_OMIT))
        yield (dirpath, dirnames, filenames)

os.walk = patched_os_walk
# END IF TEMPORARY FIX.


test_requirements = [
    'tox',
    'tox-pyenv',
    'mock',
    'factory-boy',
    'pydocstyle',
    'pytest',
    'pytest-cov',
    'pytest-flakes',
    'pytest-pep8',
    'pytest-django',
    'pytest-isort',
]


setup(
    name='django-cruditor',
    version=VERSION,
    url='https://github.com/moccu/django-cruditor',
    author='Moccu GmbH & Co. KG',
    author_email='info@moccu.com',
    description='Set of generic class based views with Bootstrap templates.',
    long_description=read('README.rst'),
    license='BSD',
    packages=find_packages(exclude=['examples*', 'tests*']),
    include_package_data=True,
    install_requires=[
        'django>=1.8,<1.10',
        'django-floppyforms>=1.6,<1.7',
        'django-filter>=0.13,<0.14',
        'django-tables2>=1.1,<1.2',
    ],
    extras_require={
        'tests': test_requirements,
        'docs': ['sphinx>=1.4,<1.5'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]
)
